import sys
import re
import os

from Subprog import Subprog
from Translator import Translator

print("ASH (Assembly Shorthand Syntax)")

location = sys.argv[1]

print(os.getcwd())

f = open(f"{os.getcwd()}/{location}", "r")
lines = f.readlines()
f.close()

code = "".join(lines)

sizes = {
    "byte": "b",
    "word": "w",
    "doubleword": "d",
    "quadword": "q",
    "b": "b",
    "w": "w",
    "d": "d",
    "q": "q",
    "dword": "d",
    "qword": "q"
}

asm = ""

sectiondatafound = re.findall(r"<.data>(.*)</.data>", code, re.DOTALL)

if len(sectiondatafound) == 1:
    print("✅ Data")
    sectiondata = sectiondatafound[0]

    asm += "section .data\n"

    for line in sectiondata.split("\n"):
        stripline = line.strip()
        vals = stripline.split(":")
        if len(vals) == 1 and stripline != "}" and stripline != "":
            asm += line.replace("{", "").strip() + ":\n"
        elif len(vals) == 2:
            asm += f"d{sizes.get(vals[0])} {vals[1]}\n"

    asm += "\n"
else:
    print("❌ Data")

sectionbssfound = re.findall(r"<.bss>(.*)</.bss>", code, re.DOTALL)

if len(sectionbssfound) == 1:
    print("✅ BSS")
    sectionbss = sectionbssfound[0]

    asm += "section .bss\n"

    for line in sectionbss.split("\n"):
        stripline = line.strip()
        vals = stripline.split(" ")
        if len(vals) == 2 and "{" in stripline:
            asm += line.replace("{", "").strip() + ":\n"
        elif len(vals) == 2:
            asm += f"res{sizes.get(vals[1])} {vals[0]}\n"

    asm += "\n"
else:
    print("❌ BSS")


sectiontextfound = re.findall(r"<.text>(.*)</.text>", code, re.DOTALL)

if len(sectiontextfound) == 1:
    print("✅ Text")
    sectiontext = sectiontextfound[0]

    asm += "section .text\n"

    subprogs = re.findall(r"<([a-zA-Z_][a-zA-Z0-9_]*)\s{(.*)?}>(.*)</\1>", sectiontext, re.DOTALL)

    main = re.findall(r"<.main>(.*)</.main>", sectiontext, re.DOTALL)

    for subprog in subprogs:
        s = Subprog(subprog[0], subprog[1].split(","), subprog[2])
        asm += s.asm()

    if len(main) == 1:
        asm += "global CMAIN\nCMAIN:\n"
        asm += "push ebp\nmov ebp, esp\n"
        asm += Translator.translate(main[0])
        asm += "pop ebp\nret\n\n"
else:
    print("❌ Text")

if asm != "":
    filename = location.replace(".ash", ".asm")
    f = open(filename, "w")
    f.write(asm);
    f.close()
    print(f"\n✅ {filename} was created successfully.")
else:
    print(f"\n❌ {location} is blank.")


class Unop:
    def __init__(self, o, op) -> None:
        self.o = o.strip()

        if "[" in self.o and "]" in self.o:
            num = int(self.o.replace("[", "").replace("]", ""))
            self.o = f"[ebp + {8 + num*4}]"

        self.op = op

    def asm(self):
        instructions = {
            "push": "push",
            "pop": "pop",
            "*": "mul",
            "/": "div",
            "!": "not",
            "printstr": "PRINT_STRING",
            "printnum": "PRINT_DEC 4,",
            "call": "call",
            "jump": "jump",
            "++": "inc",
            "--": "dec"
        }

        required = {
            "mul": ("xor edx, edx\n", ""),
            "div": ("xor edx, edx\n", "")
        }

        instr = instructions.get(self.op)
        start = ""
        end = ""
        if instr in required:
            reqs = required.get(instr)
            start = reqs[0]
            end = reqs[1]
       
        return f"{start}{instructions.get(self.op)} {self.o}\n{end}"


from Translator import Translator


class Subprog:
    def __init__(self, name, keeplist, ashcode) -> None:
        self.name = name
        self.keeplist = keeplist
        self.ashcode = ashcode
        self.code = Translator.translate(self.ashcode)

        if len(self.keeplist) == 1 and self.keeplist[0] == "":
            self.keeplist = []
    
    def asm(self):
        start = "push ebp\nmov ebp, esp\n"
        for reg in self.keeplist:
            start += f"push {reg.strip()}\n"
        
        end = "pop ebp\nret\n\n"
        for reg in self.keeplist:
            end = f"pop {reg.strip()}\n" + end 
        return f"{self.name}:\n{start}\n{self.code}\n{end}"

    


class Binop:
    def __init__(self, o1, o2, op) -> None:
        self.o1 = o1.strip()

        if "(" in self.o1 and ")" in self.o1:
            num = int(self.o1.replace("(", "").replace(")", ""))
            self.o1 = f"[ebp + {8 + num*4}]"

        self.o2 = o2.strip()

        if "(" in self.o2 and ")" in self.o2:
            num = int(self.o2.replace("(", "").replace(")", ""))
            self.o2 = f"[ebp + {8 + num*4}]"


        self.op = op

    def asm(self):
        instructions = {
            "=": "mov",
            "+=": "add",
            "-=": "sub",
            "*=": "imul",
            "\=": "div",
            "%=": "mod",
            "jump": "jump",
            "??": "cmp",
            "cmp": "cmp",
            "test": "test",
            "?": "test"
        }

        jumps = {
            "if equals": "je",
            "if zero": "jz",
            "if not equals": "jne",
            "if not zero": "jnz",
            "if less": "jl",
            "if greater": "jg"
        }

        required = {
            "div": ("xor edx, edx\n", "")
        }

        instr = instructions.get(self.op)
        start = ""
        end = ""
        if instr in required:
            reqs = required.get(instr)
            start = reqs[0]
            end = reqs[1]

        if self.op == "jump":
            return f"{jumps.get(self.o1)} {self.o2}\n"

        return f"{start}{instructions.get(self.op)} {self.o1}, {self.o2}\n{end}"


from curses.ascii import isblank
from ossaudiodev import SOUND_MIXER_TREBLE
import re

from Binop import Binop
from Unop import Unop
class Translator:
    def translate(code):
        asm = ""
        ops= re.findall(r"((.*)\s([=><\+\*-/&\|!%\?(push)(pop)(jump)(call)(cmp)(test)(printstr)(printnum)]+)\s(.*))|(<(.*)>)", code)
        for op in ops:
            if (op[5].strip() != ""):
                asm += op[5].strip() + ":\n"
            elif op[1].strip() != "":
                b = Binop(op[1], op[3], op[2])
                asm += b.asm()
            else:
                u = Unop(op[3], op[2])
                asm += u.asm()
        return asm